"""
This is a setup.py script originally generated by py2applet and then modified.

Usage:
    python setup.py py2app
"""

from setuptools import setup

from glob import glob

plist = dict(
    CFBundleName = 'Structured Annotations for 2D-to-3D Modeling',
    CFBundleDisplayName = 'Structured Annotations for 2D-to-3D Modeling',
    CFBundleIdentifier = 'com.yotamgingold.structured_annotations',
    CFBundleSignature = 'saf2',
    NSHumanReadableCopyright = 'Copyright 2008-2020 Yotam Gingold',
    CFBundleGetInfoString = '1.0.1, Copyright 2008-2020 Yotam Gingold',
    CFBundleShortVersionString = '1.0.1',
    CFBundleVersion = '1.0.1'
)

try:
    import importlib, os
    GEOMETRY_LIB = os.path.split(importlib.util.find_spec('geometry').origin)[1]
except (ImportError, AttributeError): # python 2
    import imp
    GEOMETRY_LIB = imp.find_module('geometry')[1]

APP = ['wx_annotate_collection.py']
DATA_FILES = [
 ('modes', glob('modes/*.png')),
 ('scenes/artwork', ['scenes/artwork/xy grid2.png']),
 ('ext', ['./ext/sphere4.out.obj']),
 'widgets',
 'xsections',
 
 GEOMETRY_LIB, # 'geometry.so',
 'Eline_ad.dylib',
 'edge_distances.dylib',
 'face_stitch_pattern.dylib',
 'libtritri.dylib',
 'liblaplacian_editing.dylib',
 
 'OpenSans-ExtraBold.ttf'
 ]
OPTIONS = {
    ## https://stackoverflow.com/questions/48448911/py2app-s-window-wxpython-doesn-t-come-to-front-on-launch?rq=1
    'argv_emulation': False,
    'optimize': 2,
    'graph': True,
    'excludes': ['matplotlib','zmq','sip','email','pydoc','docutils','pip','py2app','cython','distutils'],
    'no_strip': True,
    'plist': plist,
    'iconfile': 'icon/structured_annotations.icns',
    ## wx flag discussion:
    ## https://discuss.wxpython.org/t/error-with-standalone-wxpython-app-created-with-py2app/30120
    ## https://stackoverflow.com/questions/52849097/python-virtual-env-py2app-build-error-with-wxpython/54190335#54190335
    'packages': ['wx','PIL'],
    ## Prevent infinite recursion.
    ## UPDATE: I believe adding 'distutils' to the 'excludes' above is what actually fixed this.
    ## From: https://stackoverflow.com/questions/7701255/py2app-ioerror-errno-63-file-name-too-long
    ## From: https://voxpopuli.cc/2018/08/28/py2app-filename-too-long/
    # 'bdist_base': os.path.join( os.getcwd(), os.pardir, os.path.split( os.getcwd() )[1] + '-build' ),
    # 'bdist_base': os.path.join( os.getcwd(), os.pardir, 'build' ),
    # 'dist_dir': os.path.join( os.getcwd(), os.pardir, os.path.split( os.getcwd() )[1] + '-dist' )
    # 'dist_dir': os.path.join( os.getcwd(), os.pardir, 'dist' )
    }

setup(
    app=APP,
    data_files=DATA_FILES,
    options={'py2app': OPTIONS},
    setup_requires=['py2app'],
)
